/* Copyright (C)  2006-2009, Vlad Volodin <vest.84@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <glibmm/i18n.h>
#include <dirent.h>
#include <string>
#include <stdlib.h>
#include <sys/stat.h>
#include <fstream>

#ifdef NEED_PROPER_HOME_DIR
#include <unistd.h>
#include <pwd.h>
#endif //NEED_PROPER_HOME_DIR

#include "background.h"

const char Background::EXTENSION[] = ".svg";
const char Background::BACKGROUND[] = "background.";

TStringMap Background::bgMap = TStringMap ();

void Background::Clear ()
{
#ifndef NDEBUG
    std::cout << "Background::Clear" << std::endl;
#endif //NDEBUG
    Background::bgMap.clear ();
}

void Background::ScanGameDir ()
{
    ScanDir (GNOME_PIXMAPSDIR);
}

void Background::ScanHomeDir ()
{
    /* $XDG_DATA_HOME defines the base directory relative to which user
     * specific data files should be stored. If $XDG_DATA_HOME is either not
     * set or empty, a default equal to $HOME/.local/share should be used. */
    Glib::ustring homeDir = GetXDGBaseDir ("XDG_DATA_HOME", "/.local/share");

    Glib::ustring readmePath = homeDir;
    readmePath += "/README";

    std::ofstream readmeOut (readmePath.c_str ());
    readmeOut << _ ("You should place here background files. They should be saved as SVG-files\n"
                    "They should be square view for good looking. And their names must be in the\n"
                    "followed format: \'background.name.divided.with.dots.svg\'");
    readmeOut.close ();

    ScanDir (homeDir.c_str ());
}

void Background::ScanDir (const char* path) throw (file_exception)
{
#ifndef NDEBUG
    std::cout << "Background::ScanDir: " << path << std::endl;
#endif //NDEBUG
    DIR* dir = opendir (path);
    Glib::ustring fullPath = path;
    fullPath += '/';

    if (dir == NULL)
        throw file_exception (_ ("Couldn't open the current directory"), path);

    dirent* entry = NULL;
    Glib::ustring fileName;
    int posExtension = 0, posBackground = 0;
    while (entry = readdir (dir))
    {
        fileName = entry->d_name;

        posExtension = fileName.find (Background::EXTENSION);
        posBackground = fileName.find (Background::BACKGROUND);
        if ((posExtension == Glib::ustring::npos) || (posBackground != 0))
            continue;

#ifndef NDEBUG
        std::cout << fullPath + fileName << '\t' << GetTitle (fileName) << std::endl;
#endif //NDEBUG
        Background::bgMap[fullPath + fileName] = GetTitle (fileName);
    }

    closedir (dir);
}

Glib::ustring Background::GetTitle (Glib::ustring fileName)
{
    int posExtension = 0, posBackground = 0;
    posExtension = fileName.find (Background::EXTENSION);
    posBackground = fileName.find (Background::BACKGROUND);

    // Filename is "background.svg", we couldn't extract any titles
    if (posExtension == strlen (Background::BACKGROUND) - 1)
        return _ ("Unknown title");

    // extract substring between BACKGROUND and EXTENSION (without dots)
    std::string title;

    title.insert (0, fileName, strlen (Background::BACKGROUND),
                  fileName.length () - strlen (Background::BACKGROUND)
                  - strlen (Background::EXTENSION));

    title[0] = toupper (title[0]); // for good look

    // replacing dots with spaces
    int pos = 0;
    while ((pos = title.find ('.')) != std::string::npos)
        title[pos] = ' ';

    return title;
}

TStringMap Background::GetMap ()
{
    return Background::bgMap;
}

Glib::ustring Background::GetXDGBaseDir (Glib::ustring env, Glib::ustring def) throw (file_exception)
{
    // http://standards.freedesktop.org/basedir-spec/latest/index.html
    char* strDataPath = getenv (env.c_str ());

    /* The GNU C Library
     * 25.4.2 Standard Environment Variables
     * The user can set HOME to any value. If you need to make sure to obtain
     * the proper home directory for a particular user, you should not use HOME */
#ifdef NEED_PRECISION
    char* userName = getlogin ();
    passwd* pass = getpwnam (userName);
#endif //NEED_PRECISION
    Glib::ustring homeDir;
    if (strDataPath == NULL)
    {
#if NEED_PROPER_HOME_DIR
        homeDir = pass->pw_dir;
#else //NEED_PROPER_HOME_DIR
        homeDir = getenv ("HOME");
#endif //NEED_PROPER_HOME_DIR
        homeDir += def;
    }
    else
        homeDir = strDataPath;

    homeDir += "/" PACKAGE; // "/gnome-quod"

    errno = 0;
    mkdir (homeDir.c_str (), 0755);

    if ((errno != 0) && (errno != EEXIST))
        throw file_exception (_ ("Couldn't create the directory"), homeDir.c_str ());

    return homeDir;
}
