/* Copyright (C)  2006-2009, Vlad Volodin <vest.84@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <glib-2.0/glib.h>

#include "boardpanel.h"

const Glib::ustring BoardPanel::NO_BACKGROUND = "None";

BoardPanel::BoardPanel (const Board* _board) : board (_board), sx (0.0), sy (0.0),
lineWidth (3.0)
{
    set_size_request (MIN_HEIGHT, MIN_HEIGHT);

    GError* error = NULL;

    rsvgWhite = rsvg_handle_new_from_file (GNOME_PIXMAPSDIR "/white.svg", &error);
    rsvgEmpty = rsvg_handle_new_from_file (GNOME_PIXMAPSDIR "/empty.svg", &error);
    rsvgBlue = rsvg_handle_new_from_file (GNOME_PIXMAPSDIR "/blue.svg", &error);
    rsvgRed = rsvg_handle_new_from_file (GNOME_PIXMAPSDIR "/red.svg", &error);

    rsvgBackground = NULL;

    g_assert ((rsvgWhite != NULL) && (rsvgRed != NULL) &&
              (rsvgBlue != NULL) && (rsvgEmpty != NULL));

    rsvg_handle_get_dimensions (rsvgBlue, &dBlue);
    rsvg_handle_get_dimensions (rsvgRed, &dRed);
    rsvg_handle_get_dimensions (rsvgWhite, &dWhite);
    rsvg_handle_get_dimensions (rsvgEmpty, &dEmpty);

    mustRedraw = true;
    drawBackground = false;
    hasSquare = false;
    row1 = col1 = row2 = col2 = row3 = col3 = row4 = col4 = 0;
    player = RED_PLAYER;
}

void BoardPanel::Assign (const Board* _board)
{
    board = _board;
    hasSquare = false;
    row1 = col1 = row2 = col2 = row3 = col3 = row4 = col4 = 0;
    player = RED_PLAYER;
    mustRedraw = true;
}

BoardPanel::~BoardPanel ()
{
    GError* error = NULL;

    g_object_unref (rsvgWhite);
    g_object_unref (rsvgRed);
    g_object_unref (rsvgBlue);
    g_object_unref (rsvgEmpty);
    if (rsvgBackground != NULL)
        g_object_unref (rsvgBackground);
}

TCoordCell BoardPanel::PositionToCell (double x, double y) const
{
    int col = 0, row = 0;

    // GNU C Library: 20.8.3 Rounding Functions     
    // You can also convert floating-point numbers to integers simply by
    // casting them to int. This discards the fractional part, effectively
    // rounding towards zero.
    col = static_cast<int> (x / sx / (dx + dEmpty.width));
    row = static_cast<int> (y / sy / (dy + dEmpty.height));

    return TCoordCell (row, col);
}

void BoardPanel::RedrawCell (const TCoordCell& cell)
{
    // This is where we draw on the window
    Glib::RefPtr<Gdk::Window> window = get_window ();
    if (window)
    {
        RsvgHandle* rsvgQuod = NULL;

        crMemory->save ();
        crMemory->scale (sx, sy);

        if (board->getCell (cell.row, cell.col) == RED_CELL)
            rsvgQuod = rsvgRed;
        else if (board->getCell (cell.row, cell.col) == BLUE_CELL)
            rsvgQuod = rsvgBlue;
        else if (board->getCell (cell.row, cell.col) == BLOCK_CELL)
            rsvgQuod = rsvgWhite;
            //else if (board.getCell (cell.row, cell.col) == EMPTY_CELL)
        else
            rsvgQuod = rsvgEmpty;

        crMemory->save ();
        crMemory->translate (cell.col * (dx + dEmpty.width), cell.row * (dEmpty.height + dy));
        rsvg_handle_render_cairo (rsvgQuod, crMemory->cobj ());
        crMemory->restore ();

        crMemory->restore ();

        Cairo::RefPtr<Cairo::Context> cr = window->create_cairo_context ();

        cr->save ();

        queue_draw_area (cell.col * sx * (dx + dEmpty.width),
                         cell.row * sy * (dEmpty.height + dy),
                         dEmpty.width * sx, dEmpty.height * sy);

        cr->restore ();
    }
}

void BoardPanel::DrawSquare (int _row1, int _col1,
                             int _row2, int _col2,
                             int _row3, int _col3,
                             int _row4, int _col4,
                             TPlayer _player)
{
    Glib::RefPtr<Gdk::Window> window = get_window ();
    if (window)
    {
        row1 = _row1;
        col1 = _col1;
        row2 = _row2;
        col2 = _col2;
        row3 = _row3;
        col3 = _col3;
        row4 = _row4;
        col4 = _col4;
        player = _player;
        hasSquare = true;

        Gtk::Allocation allocation = get_allocation ();
        const int width = allocation.get_width ();
        const int height = allocation.get_height ();
        RsvgHandle* rsvgQuod = NULL;

        crMemory->save ();
        crMemory->scale (sx, sy);

        if (player == RED_PLAYER)
            crMemory->set_source_rgba (1.0, 0.0, 0.0, 0.8);
        else
            crMemory->set_source_rgba (0.0, 0.0, 1.0, 0.8);

        crMemory->set_line_width (lineWidth);

        crMemory->move_to (col1 * (dx + dEmpty.width) + dEmpty.width / 2,
                           row1 * (dEmpty.height + dy) + (dEmpty.height - lineWidth) / 2);
        crMemory->line_to (col2 * (dx + dEmpty.width) + dEmpty.width / 2,
                           row2 * (dEmpty.height + dy) + (dEmpty.height - lineWidth) / 2);
        crMemory->line_to (col3 * (dx + dEmpty.width) + dEmpty.width / 2,
                           row3 * (dEmpty.height + dy) + (dEmpty.height - lineWidth) / 2);
        crMemory->line_to (col4 * (dx + dEmpty.width) + dEmpty.width / 2,
                           row4 * (dEmpty.height + dy) + (dEmpty.height - lineWidth) / 2);
        crMemory->close_path ();
        crMemory->stroke ();

        crMemory->restore ();
    }
}

void BoardPanel::AssignBackground (Glib::ustring filePath)
{
    GError* error = NULL;

    drawBackground = (filePath != BoardPanel::NO_BACKGROUND);
    mustRedraw = true;

    if (rsvgBackground)
    {
        g_object_unref (rsvgBackground);
        rsvgBackground = NULL;
    }

    if (drawBackground)
    {
        rsvgBackground = rsvg_handle_new_from_file (filePath.c_str (), &error);
        rsvg_handle_get_dimensions (rsvgBackground, &dBackground);
    }

    queue_draw ();
}

void BoardPanel::SetOptions(const TOptions& options)
{
    lineWidth = options.lineWidth;
}

bool BoardPanel::on_expose_event (GdkEventExpose* event)
{
    // This is where we draw on the window

    Glib::RefPtr<Gdk::Window> window = get_window ();
    if (window)
    {
        Gtk::Allocation allocation = get_allocation ();
        const int width = allocation.get_width ();
        const int height = allocation.get_height ();

        if (mustRedraw)
        {
            imageSurface = Cairo::ImageSurface::create (Cairo::FORMAT_ARGB32, width, height);
            crMemory = Cairo::Context::create (imageSurface);

            backgroundSurface = Cairo::ImageSurface::create (Cairo::FORMAT_ARGB32, width, height);
            crbgMemory = Cairo::Context::create (backgroundSurface);
        }

#ifndef NDEBUG
        std::cout << "width = " << width << " height = " << height << std::endl;
        std::cout << "x = " << event->area.x << " y = " << event->area.y << " w = "
                << event->area.width << " h = " << event->area.height << std::endl;
#endif // NDEBUG

        // sx must be equal to sy, because of AspectFrame
        sx = width / ((dx + dEmpty.width) * Board::MAX_COL);
        sy = height / ((dy + dEmpty.height) * Board::MAX_ROW);

        sbx = static_cast<double> (width) / static_cast<double> (dBackground.width);
        sby = static_cast<double> (height) / static_cast<double> (dBackground.height);


#ifndef NDEBUG
        std::cout << sx << '\t' << sy << std::endl;
#endif // NDEBUG

        if (mustRedraw)
        {
            MakeImage ();
            mustRedraw = false;
        }

        Cairo::RefPtr<Cairo::Context> cr = window->create_cairo_context ();
        cr->save ();

        RestoreImage (cr, event->area.x, event->area.y,
                      event->area.width, event->area.height);

        cr->restore ();

        return true;
    }

    return false;
}

void BoardPanel::on_size_allocate (Gtk::Allocation& allocation)
{
#ifndef NDEBUG
    std::cout << "Boardpanel::on_size_allocate" << std::endl;
#endif // NDEBUG
    mustRedraw = true;

    Gtk::DrawingArea::on_size_allocate (allocation);
}

void BoardPanel::MakeImage ()
{
    // Draw background
    if (drawBackground)
    {
        crbgMemory->save ();
        crbgMemory->scale (sbx, sby);
#ifndef NDEBUG
        std::cout << "Boardpanel::MakeImage " << sbx << ' ' << sby << std::endl;
#endif // NDEBUG
        rsvg_handle_render_cairo (rsvgBackground, crbgMemory->cobj ());
        crbgMemory->restore ();
    }

    // Draw board
    RsvgHandle* rsvgQuod = NULL;

    crMemory->save ();
    crMemory->scale (sx, sy);

    for (int row = 0; row < Board::MAX_ROW; row++)
    {
        for (int col = 0; col < Board::MAX_COL; col++)
        {
#ifndef NDEBUG
            std::cout << "row = " << row << "\tcol = " << col << std::endl;
#endif // NDEBUG

            if (board->getCell (row, col) == RED_CELL)
                rsvgQuod = rsvgRed;
            else if (board->getCell (row, col) == BLUE_CELL)
                rsvgQuod = rsvgBlue;
            else if (board->getCell (row, col) == BLOCK_CELL)
                rsvgQuod = rsvgWhite;
                //else if (board.getCell (row, col) == EMPTY_CELL)
            else
                rsvgQuod = rsvgEmpty;

            if (Board::IsValidSquare (row, col))
                rsvg_handle_render_cairo (rsvgQuod, crMemory->cobj ());
            crMemory->translate (dx + dEmpty.width, 0);
        }
        crMemory->translate (-Board::MAX_COL * (dx + dEmpty.width), dEmpty.height + dy);
    }
    crMemory->restore ();
}

void BoardPanel::RestoreImage (Cairo::RefPtr<Cairo::Context> cr,
                               double x, double y, double width, double height)
{
    if (hasSquare)
        DrawSquare (row1, col1, row2, col2, row3, col3, row4, col4, player);
#ifndef NDEBUG
    std::cout << "BoardPanel::RestoreImage" << x << ' ' << y
            << ' ' << width << ' ' << height << std::endl;
#endif // NDEBUG

    // clip to the area indicated by the expose event so that we only redraw
    // the portion of the window that needs to be redrawn
    cr->rectangle (x, y, width, height);
    cr->set_line_width (0.0);

    cr->clip ();

    if (drawBackground)
    {
        cr->save ();
        cr->set_source (backgroundSurface, 0, 0);
        cr->paint ();
        cr->restore ();
    }

    cr->set_source (imageSurface, 0, 0);
    cr->paint ();
}

