/* Copyright (C)  2006-2009, Vlad Volodin <vest.84@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <gtkmm/main.h>
#include <glibmm/i18n.h>
#include <librsvg/rsvg.h>
#include <iostream>
#include <algorithm>

#include "boardpanel.h"

#include "types.h"
#include "mainwindow.h"
#include "background.h"
#include "options.h"
#include "optionswindow.h"
#include "statwindow.h"

MainWindow::MainWindow () :
board (new Board ()),
playerBlue (new Player (BLUE_PLAYER)),
playerRed (new Player (RED_PLAYER)),
currentState (END_STATE),
playerAI (NULL),
blueAI (NULL),
redAI (NULL),
allowInput (false),
m_mainBox (false, 5),
m_rightPlayerBox (false, 5),
m_bluePanel (playerBlue),
m_redPanel (playerRed),
m_blueLabel (_ ("Blue Player")),
m_redLabel (_ ("Red Player")),
m_boardPanel (board)
{
    set_title ("Gnome Quod");
    set_default_size (DEFAULT_WIDTH, DEFAULT_HEIGHT);

    newOptions = TOptions::LoadOptions ();
    statistics = TStatistics::LoadStatistics ();

    try
    {
        this ->set_icon_from_file (GNOME_ICONDIR "/quod.png");
    }
    catch (Glib::Exception & e)
    {
        std::cout << e .what () << std::endl;
    }

    add (m_mainBox);

    Gtk::Widget* pMenubar = BuildMenu ();
    if (pMenubar)
    {
        m_mainBox.pack_start (*pMenubar, Gtk::PACK_SHRINK);
        pMenubar->show ();
    }

    m_mainBox.pack_start (m_gameBox);
    m_mainBox.pack_end (m_statusBox, Gtk::PACK_SHRINK);
    m_statusBox.pack_start (m_bluePlayerStatusBar); // left Status Bar (without triangle)
    m_statusBox.pack_start (m_redPlayerStatusBar);
    m_bluePlayerStatusBar.set_has_resize_grip (false);

    m_boardPanel.set_events (Gdk::BUTTON_PRESS_MASK);
    m_boardPanel.signal_button_press_event ().connect (sigc::mem_fun (*this,
                                                                      &MainWindow::on_board_button_press_event));
    m_boardPanel.SetOptions (newOptions);

    m_gameBox.pack_start (m_boardFrame);
    m_boardFrame.add (m_boardPanel);
    m_boardFrame.unset_label ();

    m_gameBox.pack_end (m_rightPlayerBox, Gtk::PACK_SHRINK, 5);
    m_rightPlayerBox.pack_start (m_bluePlayerBox);
    m_rightPlayerBox.pack_start (m_redPlayerBox);

    m_blueLabel.set_width_chars (15);
    m_redLabel.set_width_chars (15);
    m_bluePlayerBox.pack_start (m_blueLabel, Gtk::PACK_SHRINK);
    m_bluePlayerBox.pack_start (m_bluePanel);
    m_redPlayerBox.pack_start (m_redLabel, Gtk::PACK_SHRINK);
    m_redPlayerBox.pack_start (m_redPanel);

    NewGame ();

    show_all_children ();
}

MainWindow::~MainWindow ()
{
    options = newOptions;
    TOptions::SaveOptions (options);
    TStatistics::SaveStatistics (statistics);

    delete board;
    delete playerRed;
    delete playerBlue;
    delete blueAI;
    delete redAI;
}

bool MainWindow::on_board_button_press_event (GdkEventButton* event)
{
    if (event->type == GDK_BUTTON_PRESS && currentState != END_STATE && allowInput)
    {
#ifndef NDEBUG
        std::cout << "x = " << event->x << " y = " << event->y << std::endl;
#endif // NDEBUG

        TCoordCell cell;
        cell = m_boardPanel.PositionToCell (event->x, event->y);
#ifndef NDEBUG
        std::cout << "row = " << cell.row << " col = " << cell.col << std::endl;
#endif // NDEBUG

        if (!Board::IsValidSquare (cell))
            return false;

        if (board->getCell (cell.row, cell.col) != EMPTY_CELL)
            return false;

        if (event->button == 3)
        {
#ifndef NDEBUG
            std::cout << "Panel Right Click handled" << std::endl;
#endif // NDEBUG
            if (currentPlayer->getBlockLeft () <= 0)
                return false;

            currentPlayer->DecreaseBlock ();
            m_rightPlayerBox.queue_draw ();
            UpdatePlayerStatus ();
            board->setCell (cell.row, cell.col, BLOCK_CELL);

            m_boardPanel.RedrawCell (cell);
        }
        else if (event->button == 1)
        {
#ifndef NDEBUG
            std::cout << "Panel Left Click handled" << std::endl;
#endif // NDEBUG
            if (currentPlayer->getAttackLeft () <= 0)
            {
                EndGame ();
                return false;
            }

            currentPlayer->DecreaseAttack ();
            m_rightPlayerBox.queue_draw ();
            UpdatePlayerStatus ();

            board->setCell (cell.row, cell.col,
                            (Board::PlayerToCell (currentPlayer->getPlayer ())));

            m_boardPanel.RedrawCell (cell);
            CheckWinner ();

            currentPlayer = (currentPlayer == playerRed ? playerBlue : playerRed);
        }

        StartAI ();

        return true; //It has been handled.
    }
    else
        return false;
}

void MainWindow::on_menu_game_quit ()
{
    hide (); //Closes the main window to stop the Gtk::Main::run().
}

void MainWindow::on_menu_game_new ()
{
    if (currentState != END_STATE)
    {
        //Tell the user:
        Gtk::MessageDialog dialog (*this,
                                   _ ("Do you want to start a new game?"),
                                   false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO);
        int result = dialog.run ();

        if (result == Gtk::RESPONSE_NO)
            return;
    }

    NewGame ();
}

void MainWindow::on_menu_game_stat ()
{
    StatWindow statDialog(*this, statistics);

    statDialog.run ();
}

void MainWindow::on_menu_edit_options ()
{
    OptionsWindow optionsDialog (*this, newOptions);

    int result = optionsDialog.run ();
    if (result == Gtk::RESPONSE_OK)
    {
        optionsDialog.GetOptions (newOptions);

        return;
    }
}

void MainWindow::on_menu_help_about ()
{
    Gtk::AboutDialog dialog;

    GError* error = NULL;
    Glib::RefPtr<Gdk::Pixbuf> pbufLogo = Gdk::Pixbuf::create_from_file(GNOME_PIXMAPSDIR "/picture.png");

    dialog.set_transient_for (*this);

    dialog.set_program_name ("Gnome Quod");
    dialog.set_version (VERSION);
    dialog.set_logo (pbufLogo);
    dialog.set_copyright (_ ("Copyright © 2006-2008 Vlad Volodin (Vest)"));
    dialog.set_comments (_ ("The goal is to place pieces on a grid\nso that they form a square."));

    dialog.set_website_label (_ ("gQuod website"));
    //the url is not clickable because this would introduce an new dep (gnome-vfsmm) 
    dialog.set_website ("http://sourceforge.net/projects/gquod");

    std::list<Glib::ustring> list_authors;
    list_authors.push_back (_ ("G. Keith Still (original game's inventor)"));
    list_authors.push_back (_ ("Michael J. Mefford (Quod Delphi source, 1996)"));
    list_authors.push_back (_ ("Jan Bobrowski (VB Script) <jb@wizard.ae.krakow.pl, 1998-2002 GPL>"));
    list_authors.push_back (_ ("Vlad Volodin (Palm and Linux version) <vest.84@gmail.com>"));

    dialog.set_authors (list_authors);

    Glib::ustring license = _ ("Gnome Quod is free software; you can redistribute it and/or modify it\n"
                               "under the terms of the GNU General Public Licence as published\n"
                               "by the Free Software Foundation; either version 3 of the Licence,\n"
                               "or (at your option) any later version.\n"
                               "\n"
                               "Gnome Quod is distributed in the hope that it will be useful,\n"
                               "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
                               "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
                               "GNU General Public License for more details.\n"
                               "\n"
                               "You should have received a copy of the GNU General Public\n"
                               "Licence along with Gnome Quod; if not, write to the\n"
                               "Free Software Foundation, Inc.\n"
                               "59 Temple Place, Suite 330\n"
                               "Boston, MA  02111-1307  USA");
    dialog.set_license (license);

    /*TO TRANSLATORS: your name(s) here please, if there are more translators put newlines (\n) between the names.
      It's a good idea to provide the url of your translationteam as well. Thanks! */
    Glib::ustring str_credits = _ ("translator-credits");
    if (str_credits != "translator-credits")
        dialog.set_translator_credits (str_credits);

    dialog.run ();
}

Gtk::Widget* MainWindow::BuildMenu ()
{
    // Create actions for menus and toolbars:
    m_refActionGroup = Gtk::ActionGroup::create ("UIMainMenu");

    // File menu
    m_refActionGroup->add (Gtk::Action::create ("GameMenu", _ ("_File")));
    // File|Sub-menu
    m_refActionGroup->add (Gtk::Action::create ("GameNew", Gtk::Stock::NEW),
                           sigc::mem_fun (*this, &MainWindow::on_menu_game_new));
    m_refActionGroup->add (Gtk::Action::create ("GameStat", Gtk::Stock::INFO, _("_Statistics")),
                           Gtk::AccelKey ("<control>S"),
                           sigc::mem_fun (*this, &MainWindow::on_menu_game_stat));
    m_refActionGroup->add (Gtk::Action::create ("GameQuit", Gtk::Stock::QUIT),
                           sigc::mem_fun (*this, &MainWindow::on_menu_game_quit));

    // Edit menu
    m_refActionGroup->add (Gtk::Action::create ("EditMenu", _ ("_Edit")));
    // Edit|Sub-menu
    // Edit|Background|Sub-menu
    m_refActionGroup->add (Gtk::Action::create ("EditBgMenu", _ ("_Background")));
    m_refPanelToggle = Gtk::ToggleAction::create ("EditTogglePlayer", _ ("_Player panel"),
                                                  _ ("Show right players' panel"), true);
    m_refActionGroup->add (m_refPanelToggle, Gtk::AccelKey ("<control>P"),
                           sigc::mem_fun (*this, &MainWindow::on_menu_edit_show));
    m_refActionGroup->add (Gtk::Action::create ("EditOptions", Gtk::Stock::PREFERENCES),
                           sigc::mem_fun (*this, &MainWindow::on_menu_edit_options));

    // Help menu
    m_refActionGroup->add (Gtk::Action::create ("HelpMenu", _ ("_Help")));
    // Help|Sub-menu
    m_refActionGroup->add (Gtk::Action::create ("HelpAbout", Gtk::Stock::ABOUT),
                           Gtk::AccelKey ("<control>A"),
                           sigc::mem_fun (*this, &MainWindow::on_menu_help_about));

    m_refUIManager = Gtk::UIManager::create ();
    m_refUIManager->insert_action_group (m_refActionGroup);

    add_accel_group (m_refUIManager->get_accel_group ());

    //Layout the actions in a menubar and toolbar:
    Glib::ustring ui_info =
            "<ui>"
            "  <menubar name='MenuBar'>"
            "    <menu action='GameMenu'>"
            "      <menuitem action='GameNew'/>"            
            "      <separator/>"
            "      <menuitem action='GameStat'/>"
            "      <separator/>"
            "      <menuitem action='GameQuit'/>"
            "    </menu>"
            "    <menu action='EditMenu'>"
            "      <menu action='EditBgMenu'>"
            "      </menu>"
            "      <menuitem action='EditTogglePlayer'/>"
            "      <separator/>"
            "      <menuitem action='EditOptions'/>"
            "    </menu>"
            "    <menu action='HelpMenu'>"
            "      <menuitem action='HelpAbout'/>"
            "    </menu>"
            "  </menubar>"
            "</ui>";

#ifdef GLIBMM_EXCEPTIONS_ENABLED
    try
    {
        m_refUIManager->add_ui_from_string (ui_info);
    }
    catch (const Glib::Error& ex)
    {
        std::cerr << "building menus failed: " << ex.what ();
    }
#else
    std::auto_ptr<Glib::Error> ex;
    m_refUIManager->add_ui_from_string (ui_info, ex);
    if (ex.get ())
        std::cerr << "building menus failed: " << ex->what () << std::endl;
#endif //GLIBMM_EXCEPTIONS_ENABLED

    // Right justify the "Help" Menu, depends on TOptions
    Gtk::MenuItem* pHelpMenuItem = dynamic_cast<Gtk::MenuItem*> (m_refUIManager->get_widget ("/ui/MenuBar/HelpMenu"));
    pHelpMenuItem->set_right_justified (newOptions.helpAlign == TOptions::Right);

    //Get the menubar and toolbar widgets, and add them to a container widget:
    Gtk::Widget* pWidgetMenuBar = m_refUIManager->get_widget ("/ui/MenuBar");

    BuildBgSubmenu ();

    return pWidgetMenuBar;
}

void MainWindow::BuildBgSubmenu ()
{
    // Create actions for menus and toolbars:
    m_refActionBgGroup = Gtk::ActionGroup::create ("UIBackground");
    Gtk::RadioAction::Group group_background;

    TStringMap bgMap = Background::GetMap ();

    //Layout the actions in a menubar (Background submenu):
    Glib::ustring ui_info =
            "<ui>"
            "  <menubar name='MenuBar'>"
            "    <menu action='EditMenu'>"
            "      <menu action='EditBgMenu'>"
            "        <menuitem action='EditBg0'/>"
            "        <separator/>";

    // Passes "None" as filePath, means that we don't want to have any background
    m_refActionBgGroup->add (Gtk::RadioAction::create (group_background, "EditBg0",
                                                       _ ("N_one"), _ ("None")),
                             Gtk::AccelKey ("<control>O"),
                             sigc::bind<-1 > (sigc::mem_fun (*this,
                                                             &MainWindow::on_menu_edit_background),
                             BoardPanel::NO_BACKGROUND));

    Glib::ustring label;
    int num = 1;

    Glib::ustring menuPath;

    for (TStringMap::iterator i = bgMap.begin (); i != bgMap.end (); i++, num++)
    {
        label = Glib::ustring::compose ("EditBg%1", num);

        // I use this variable, because I need to know, should I display background or not
        Glib::RefPtr<Gtk::RadioAction> refRadioAction = Gtk::RadioAction::create (group_background,
                                                                                  label,
                                                                                  i->second);
        m_refActionBgGroup->add (refRadioAction,
                                 sigc::bind<-1 > (sigc::mem_fun (*this, &MainWindow::on_menu_edit_background),
                                 i->first));

        if (newOptions.bgPath == i->first)
            refRadioAction->set_active (); // the most elegant solution :)

        ui_info += Glib::ustring::compose ("        <menuitem action='EditBg%1'/>", num);
    }

    ui_info +=
            "      </menu>"
            "    </menu>"
            "  </menubar>"
            "</ui>";

    m_refUIBgManager = Gtk::UIManager::create ();
    m_refUIBgManager->insert_action_group (m_refActionBgGroup);

    int merge_id;

#ifdef GLIBMM_EXCEPTIONS_ENABLED
    try
    {
        merge_id = m_refUIManager->add_ui_from_string (ui_info);
    }
    catch (const Glib::Error& ex)
    {
        std::cerr << "building menus failed: " << ex.what ();
    }
#else
    std::auto_ptr<Glib::Error> ex;
    merge_id = m_refUIManager->add_ui_from_string (ui_info, ex);
    if (ex.get ())
        std::cerr << "Building menus failed: " << ex->what () << std::endl;
#endif //GLIBMM_EXCEPTIONS_ENABLED

    m_refUIManager->insert_action_group (m_refActionBgGroup);
}

void MainWindow::NewGame ()
{
#ifndef NDEBUG
    std::cout << "New Game!" << std::endl;
#endif // NDEBUG    

    delete board;
    delete playerRed;
    delete playerBlue;
    delete blueAI;
    delete redAI;

    // This is preventing from fast quod placing (if BOTH_AI)
    if (aiConnection.connected ())
        aiConnection.disconnect ();

    board = NULL;
    playerRed = playerBlue = NULL;
    blueAI = redAI = NULL;

    board = new Board ();
    playerBlue = new Player (BLUE_PLAYER);
    playerRed = new Player (RED_PLAYER);

    options = newOptions;

    switch (options.firstPlayer)
    {
        case RED_PLAYER:
            currentPlayer = playerRed;
            break;
        case BLUE_PLAYER:
        default:
            currentPlayer = playerBlue;
            break;
    }

    if (options.redPlayer == EASYAI_OPTION)
        redAI = new AIEasy (*playerRed, *board);
    else if (options.redPlayer == HARDAI_OPTION)
        redAI = new AIHard (*playerRed, *board);

    if (options.bluePlayer == EASYAI_OPTION)
        blueAI = new AIEasy (*playerBlue, *board);
    else

        if (options.bluePlayer == HARDAI_OPTION)
        blueAI = new AIHard (*playerBlue, *board);

    playerAI = NULL;
    allowInput = true;
    currentState = PLAY_STATE;

    m_bluePanel.Assign (playerBlue);
    m_redPanel.Assign (playerRed);
    m_boardPanel.Assign (board);

    m_rightPlayerBox.queue_draw ();
    m_boardPanel.queue_draw ();
    UpdatePlayerStatus ();

    StartAI ();
}

void MainWindow::EndGame ()
{
    currentState = END_STATE;
    currentPlayer = NULL;

    // This is preventing from fast quod placing (if BOTH_AI)

    if (aiConnection.connected ())
        aiConnection.disconnect ();
}

void MainWindow::CheckWinner ()
{
    int row1, col1,
            row2, col2,
            row3, col3,
            row4, col4;
    TWinnerState state = NONE_WINNER;

    TCell currentPlayerCell = Board::PlayerToCell (currentPlayer->getPlayer ());

    bool hasWinner = board->FindSquare (row1, col1, row2, col2, row3, col3, row4, col4,
                                        currentPlayerCell);

    if (hasWinner)
    {
        m_boardPanel.DrawSquare (row1, col1, row2, col2, row3, col3, row4, col4,
                                 currentPlayer->getPlayer ());

        state = (currentPlayer->getPlayer () == BLUE_PLAYER ? BLUE_WINNER :
                (currentPlayer->getPlayer () == RED_PLAYER ? RED_WINNER : NONE_WINNER));

        EndGame ();
        m_boardPanel.queue_draw ();
    }
    else
    {
        if ((playerBlue->getAttackLeft () == 0) && (playerRed->getAttackLeft () == 0))
        {
            if (playerBlue->getBlockLeft () > playerRed->getBlockLeft ())
                state = BLUE_WINNER;
            else if (playerBlue->getBlockLeft () < playerRed->getBlockLeft ())
                state = RED_WINNER;
            else
                state = DRAW_WINNER;

            EndGame ();
        }
    }

    if (state == NONE_WINNER)
        return;

    statistics.IncreaseStatistics(state, options);

    Glib::ustring congratulation;
    if (state == DRAW_WINNER)
        congratulation = _ ("Draw");
    else if (state == RED_WINNER)
        congratulation = _ ("Red player won");
    else if (state == BLUE_WINNER)
        congratulation = _ ("Blue player won");
    else
        return; // it is impossible

    Gtk::MessageDialog dialog (*this, congratulation, false,
                               Gtk::MESSAGE_INFO, Gtk::BUTTONS_OK, true);

    dialog.run ();
}

void MainWindow::StartAI ()
{
    if (currentState == END_STATE)
        return;

    if ((options.bluePlayer == HUMAN_OPTION) && (currentPlayer->getPlayer () == BLUE_PLAYER))
        return;

    if ((options.redPlayer == HUMAN_OPTION) && (currentPlayer->getPlayer () == RED_PLAYER))
        return;

#ifndef NDEBUG
    std::cout << "AI Started " << std::endl;
#endif // NDEBUG   

    playerAI = (currentPlayer->getPlayer () == RED_PLAYER ?
            dynamic_cast<AIBase*> (redAI) : dynamic_cast<AIBase*> (blueAI));
    allowInput = false;

    aiConnection = Glib::signal_timeout ().connect_seconds (sigc::mem_fun (*this,
                                                                           &MainWindow::on_timeout),
                                                            options.pause);

#ifndef NDEBUG
    std::cout << "AI Stopped " << std::endl;
#endif // NDEBUG
}

bool MainWindow::on_timeout ()
{
    TMove move = NONE_MOVE;

    move = playerAI->Step ();

    TCoordCell cell = playerAI->getLastMove ();
#ifndef NDEBUG
    std::cout << "Last Step = " << cell.row << ", " << cell.col << std::endl;
#endif // NDEBUG

    if (move == BLOCK_MOVE)
    {
        currentPlayer->DecreaseBlock ();

        board->setCell (cell.row, cell.col, BLOCK_CELL);

        m_boardPanel.RedrawCell (cell);
        m_rightPlayerBox.queue_draw ();
        UpdatePlayerStatus ();
    }
    else
    {
        currentPlayer->DecreaseAttack ();

        board->setCell (cell.row, cell.col,
                        (Board::PlayerToCell (currentPlayer->getPlayer ())));

        m_boardPanel.RedrawCell (cell);
        m_rightPlayerBox.queue_draw ();
        UpdatePlayerStatus ();

        CheckWinner ();

        currentPlayer = (currentPlayer == playerRed ? playerBlue : playerRed);

        allowInput = true;
    }

    if (move == ATTACK_MOVE)
        StartAI ();

    return (move == BLOCK_MOVE);
}

void MainWindow::UpdatePlayerStatus ()
{
    m_redPlayerStatusBar.pop ();
    m_bluePlayerStatusBar.pop ();

    gchar* msgBlue = g_strdup_printf (_ ("Blue: %d / %d"),
                                      playerBlue->getAttackLeft (), playerBlue->getBlockLeft ());
    gchar* msgRed = g_strdup_printf (_ ("Red: %d / %d"),
                                     playerRed->getAttackLeft (), playerRed->getBlockLeft ());

    m_bluePlayerStatusBar.push (msgBlue);
    m_redPlayerStatusBar.push (msgRed);

    g_free (msgRed);
    g_free (msgBlue);
}

void MainWindow::on_menu_edit_show ()
{
    int width, height;
    get_size (width, height);

    if (m_refPanelToggle->get_active ())
    {
        m_rightPlayerBox.show ();
        resize (width + m_rightPlayerBox.get_width () + m_gameBox.get_spacing (), height);
    }
    else
    {
        m_rightPlayerBox.hide ();
        resize (width - m_rightPlayerBox.get_width () - m_gameBox.get_spacing (), height);
    }
}

void MainWindow::on_menu_edit_background (Glib::ustring filePath)
{
#ifndef NDEBUG
    std::cout << "MainWindow::on_menu_edit_background: " << filePath << std::endl;
#endif // NDEBUG

    options.bgPath = newOptions.bgPath = filePath;

    m_boardPanel.AssignBackground (filePath);
}
