/* Copyright (C)  2006-2009, Vlad Volodin <vest.84@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <libxml++/libxml++.h>
#include <iostream>

#include "options.h"
#include "background.h"
#include "boardpanel.h"

const Glib::ustring TOptions::FILENAME_OPTIONS = "options.xml";
const Glib::ustring TOptions::FILENAME_DTD = "options.dtd";

TOptions::TOptions () : firstPlayer (BLUE_PLAYER), bluePlayer (HUMAN_OPTION),
redPlayer (HARDAI_OPTION), bgPath (BoardPanel::NO_BACKGROUND),
pause (1), helpAlign (Left), lineWidth (3.0) { }

TOptions::TOptions (const TOptions& o)
{
    firstPlayer = o.firstPlayer;
    bluePlayer = o.bluePlayer;
    redPlayer = o.redPlayer;
    pause = o.pause;
    bgPath = o.bgPath;
    helpAlign = o.helpAlign;
    lineWidth = o.lineWidth;
}

TOptions::~TOptions () { }

TOptions& TOptions::operator = (const TOptions& o)
{
    firstPlayer = o.firstPlayer;
    bluePlayer = o.bluePlayer;
    redPlayer = o.redPlayer;
    pause = o.pause;
    bgPath = o.bgPath;
    helpAlign = o.helpAlign;
    lineWidth = o.lineWidth;

    return *this;
}

TOptions TOptions::LoadOptions ()
{
    /* $XDG_CONFIG_HOME defines the base directory relative to which user
     * specific configuration files should be stored. If $XDG_CONFIG_HOME is
     * either not set or empty, a default equal to $HOME/.config should be
     * used. */

    Glib::ustring configDir = Background::GetXDGBaseDir ("XDG_CONFIG_HOME", "/.config");

    TOptions options;

#ifdef LIBXMLCPP_EXCEPTIONS_ENABLED
    try
    {
#endif //LIBXMLCPP_EXCEPTIONS_ENABLED
        xmlpp::DomParser parser;
        parser.set_validate (); // By default, the parser will not validate the XML file.
        parser.set_substitute_entities (); //We just want the text to be resolved/unescaped automatically.
        parser.parse_file (configDir + '/' + TOptions::FILENAME_OPTIONS);
        if (parser)
        {
            //Walk the tree:
            const xmlpp::Node* pNode = parser.get_document ()->get_root_node (); //deleted by DomParser.

            if (pNode->get_name () == "Options")
            {
                xmlpp::Node::NodeList list = pNode->get_children ();
                for (xmlpp::Node::NodeList::iterator iter = list.begin (); iter != list.end (); ++iter)
                {
                    xmlpp::Node* node = *iter;

                    const xmlpp::Element* nodeElement = dynamic_cast<const xmlpp::Element*> (node);

                    if (nodeElement)
                    {
                        Glib::ustring nodeName = nodeElement->get_name (),
                                value;

                        if (nodeName == "FirstPlayer")
                        {
                            value = nodeElement->get_attribute_value ("id");

                            if (value == "Blue")
                                options.firstPlayer = BLUE_PLAYER;
                            else if (value == "Red")
                                options.firstPlayer = RED_PLAYER;
                        }
                        else if (nodeName == "BluePlayer")
                        {
                            value = nodeElement->get_attribute_value ("id");

                            if (value == "Human")
                                options.bluePlayer = HUMAN_OPTION;
                            if (value == "Easy")
                                options.bluePlayer = EASYAI_OPTION;
                            if (value == "Hard")
                                options.bluePlayer = HARDAI_OPTION;
                        }
                        else if (nodeName == "RedPlayer")
                        {
                            value = nodeElement->get_attribute_value ("id");

                            if (value == "Human")
                                options.redPlayer = HUMAN_OPTION;
                            if (value == "Easy")
                                options.redPlayer = EASYAI_OPTION;
                            if (value == "Hard")
                                options.redPlayer = HARDAI_OPTION;
                        }
                        else if (nodeName == "Pause")
                        {
                            const xmlpp::TextNode* nodeText = nodeElement->get_child_text ();

                            options.pause = g_ascii_strtoll (nodeText->get_content ().c_str (), NULL, 10);
                            if (options.pause <= 0)
                                options.pause = 1;
                        }
                        else if (nodeName == "Background")
                            options.bgPath = nodeElement->get_attribute_value ("src");
                        else if (nodeName == "HelpMenu")
                            options.helpAlign = (nodeElement->get_attribute_value ("align") == "Right" ? TOptions::Right : TOptions::Left);
                        else if (nodeName == "LineWidth")
                        {
                            const xmlpp::TextNode* nodeText = nodeElement->get_child_text ();

                            options.lineWidth = g_ascii_strtod (nodeText->get_content ().c_str (), NULL);
                            if (options.lineWidth <= 0.0)
                                options.lineWidth = 3.0;
                        }
                    }
                }
            }
        }
#ifdef LIBXMLCPP_EXCEPTIONS_ENABLED
    }
    catch (const std::exception& ex)
    {
        std::cout << "Exception caught: " << ex.what () << std::endl;
    }
#endif //LIBXMLCPP_EXCEPTIONS_ENABLED
    return options;
}

void TOptions::SaveOptions (const TOptions& options)
{
    /* $XDG_CONFIG_HOME defines the base directory relative to which user
     * specific configuration files should be stored. If $XDG_CONFIG_HOME is
     * either not set or empty, a default equal to $HOME/.config should be
     * used. */

    Glib::ustring configDir = Background::GetXDGBaseDir ("XDG_CONFIG_HOME", "/.config");

    xmlpp::Document document;
    document.set_internal_subset ("Options", "", GNOME_DATADIR "/" + TOptions::FILENAME_DTD);
    //document.set_internal_subset ("Options", "", TOptions::FILENAME_DTD);

    //Quod is the default namespace prefix.
    //Declares the namespace and uses its prefix for this node
    xmlpp::Element* nodeRoot = document.create_root_node ("Options", "http://sourceforge.net/projects/gquod", "Quod");
    nodeRoot->set_namespace_declaration ("http://sourceforge.net/projects/gquod", "Quod"); // Also associate this prefix with this namespace:
    nodeRoot->set_namespace ("Quod");

    xmlpp::Element* nodeChild = nodeRoot->add_child ("FirstPlayer");

    nodeChild->set_attribute ("id", TPLAYER_STRING[options.firstPlayer]);

    nodeChild = nodeRoot->add_child ("BluePlayer");
    nodeChild->set_attribute ("id", TPLAYEROPTION_STRING[options.bluePlayer]);

    nodeChild = nodeRoot->add_child ("RedPlayer");
    nodeChild->set_attribute ("id", TPLAYEROPTION_STRING[options.redPlayer]);

    nodeRoot->add_child_comment ("Pause: Seconds between each AI movement");
    nodeChild = nodeRoot->add_child ("Pause");
    nodeChild->set_child_text (Glib::ustring::format (options.pause));

    nodeChild = nodeRoot->add_child ("Background");
    nodeChild->set_attribute ("src", options.bgPath);

    if (options.helpAlign == TOptions::Right)
    {
        nodeChild = nodeRoot->add_child ("HelpMenu");
        nodeChild->set_attribute ("align", "Right");
    }

    nodeChild = nodeRoot->add_child ("LineWidth");
    nodeChild->set_child_text (Glib::ustring::format (options.lineWidth));

#ifndef NDEBUG
    Glib::ustring whole = document.write_to_string_formatted ();

    std::cout << "XML built at runtime: " << std::endl << whole << std::endl;
    std::cout << "default namespace: " << nodeRoot->get_namespace_uri () << std::endl;
#else //NDEBUG
    document.write_to_file_formatted (configDir + '/' + TOptions::FILENAME_OPTIONS);
#endif //NDEBUG
}
