/* Copyright (C)  2006-2009, Vlad Volodin <vest.84@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <libxml++/libxml++.h>

#include "statistics.h"
#include "background.h"
#include "options.h"
#include "types.h"

const Glib::ustring TPlayerStat::GAME_TYPE[] = {"HumanHuman", "HumanEasy", "HumanHard",
    "EasyEasy", "EasyHard", "HardHard"};

TPlayerStat::TPlayerStat () : human_human (0), human_easy (0), human_hard (0),
easy_easy (0), easy_hard (0), hard_hard (0) { }

TPlayerStat::TPlayerStat (const TPlayerStat& o)
{
    human_human = o.human_human;
    human_easy = o.human_easy;
    human_hard = o.human_hard;
    easy_easy = o.easy_easy;
    easy_hard = o.easy_hard;
    hard_hard = o.hard_hard;
}

TPlayerStat& TPlayerStat::operator = (const TPlayerStat& o)
{
    if (&o == this) return *this;

    human_human = o.human_human;
    human_easy = o.human_easy;
    human_hard = o.human_hard;
    easy_easy = o.easy_easy;
    easy_hard = o.easy_hard;
    hard_hard = o.hard_hard;

    return *this;
}

TPlayerStat::~TPlayerStat () { }

TStatistics::TStatistics () { }

TStatistics::TStatistics (const TStatistics& o)
{
    blue_blue = o.blue_blue;
    blue_red = o.blue_red;
    red_blue = o.red_blue;
    red_red = o.red_red;
}

TStatistics& TStatistics::operator = (const TStatistics& o)
{
    if (&o == this) return *this;

    blue_blue = o.blue_blue;
    blue_red = o.blue_red;
    red_blue = o.red_blue;
    red_red = o.red_red;

    return *this;
}

const Glib::ustring TStatistics::FILENAME_STATISTICS = "statistics.xml";
const Glib::ustring TStatistics::FILENAME_DTD = "statistics.dtd";

TStatistics::~TStatistics () { }

void TStatistics::IncreaseStatistics (TWinnerState state, TOptions curOptions)
{
    if (state == NONE_WINNER)
        return;

    TPlayerStat* plStat = NULL;

    if (state == BLUE_WINNER)
        plStat = (curOptions.firstPlayer == BLUE_PLAYER ? &blue_blue : &red_blue);
    else if (state == RED_WINNER)
        plStat = (curOptions.firstPlayer == BLUE_PLAYER ? &blue_red : &red_red);
    else if (state == DRAW_WINNER)
        plStat = (curOptions.firstPlayer == BLUE_PLAYER ? &blue_draw : &red_draw);

    if (curOptions.bluePlayer == HUMAN_OPTION)
    {
        if (curOptions.redPlayer == HUMAN_OPTION)
            plStat->human_human++;
        else if (curOptions.redPlayer == EASYAI_OPTION)
            plStat->human_easy++;
        else if (curOptions.redPlayer == HARDAI_OPTION)
            plStat->human_hard++;
    }
    else if (curOptions.bluePlayer == EASYAI_OPTION)
    {
        if (curOptions.redPlayer == EASYAI_OPTION)
            plStat->easy_easy++;
        else if (curOptions.redPlayer == HARDAI_OPTION)
            plStat->easy_hard++;
    }
    else if (curOptions.bluePlayer == HARDAI_OPTION)
    {
        if (curOptions.redPlayer == HARDAI_OPTION)
            plStat->hard_hard++;
    }
}

TStatistics TStatistics::LoadStatistics ()
{
    /* $XDG_CONFIG_HOME defines the base directory relative to which user
     * specific configuration files should be stored. If $XDG_CONFIG_HOME is
     * either not set or empty, a default equal to $HOME/.config should be
     * used. */

    Glib::ustring homeDir = Background::GetXDGBaseDir ("XDG_DATA_HOME", "/.local/share");

    TStatistics statistics;

#ifdef LIBXMLCPP_EXCEPTIONS_ENABLED
    try
    {
#endif //LIBXMLCPP_EXCEPTIONS_ENABLED
        xmlpp::DomParser parser;
        parser.set_validate (); // By default, the parser will not validate the XML file.
        parser.set_substitute_entities (); //We just want the text to be resolved/unescaped automatically.
        parser.parse_file (homeDir + '/' + TStatistics::FILENAME_STATISTICS);

        if (parser)
        {
            //Walk the tree:
            const xmlpp::Node* pNode = parser.get_document ()->get_root_node (); //deleted by DomParser.

            if (pNode->get_name () == "Statistics")
            {
                xmlpp::Node::NodeList firstPlayerList = pNode->get_children ();
                for (xmlpp::Node::NodeList::iterator iter = firstPlayerList.begin ();
                        iter != firstPlayerList.end (); ++iter)
                {
                    xmlpp::Node* node = *iter;
                    const xmlpp::Element* nodeElement = dynamic_cast<const xmlpp::Element*> (node);

                    TPlayerStat * blue[] = {&statistics.blue_blue, &statistics.blue_red, &statistics.blue_draw},
                    * red[] = {&statistics.red_blue, &statistics.red_red, &statistics.red_draw};

                    if (nodeElement)
                    {
                        Glib::ustring nodeName = nodeElement->get_name ();
                        if (nodeName == "FirstPlayer")
                        {
                            Glib::ustring value = nodeElement->get_attribute_value ("id");
                            TPlayerStat** pPlayer = NULL;
                            if (value == "Blue")
                                pPlayer = blue;
                            else if (value == "Red")
                                pPlayer = red;
                            else
                                continue;

                            xmlpp::Node::NodeList gameTypeList = nodeElement->get_children ();
                            for (xmlpp::Node::NodeList::iterator jiter = gameTypeList.begin ();
                                    jiter != gameTypeList.end (); ++jiter)
                            {
                                xmlpp::Node* pGameType = *jiter;

                                if (pGameType->get_name () == "GameType")
                                {
                                    const xmlpp::Element* gameTypeNode = dynamic_cast<const xmlpp::Element*> (pGameType);
                                    const xmlpp::TextNode* nodeText = gameTypeNode->get_child_text ();

                                    Glib::ustring gameTypeValue = gameTypeNode->get_attribute_value ("id");
                                    char colon;
                                    std::istringstream str(nodeText->get_content ());

                                    // sscanf(nodeText->get_content ().c_str(), "%d:%d:%d", &a, &b, &c);
                                    // or I use "char colon;" to drop ':'
                                    // thanks waqas from c++@conference.jabber.ru
                                    if (gameTypeValue == "HumanHuman")
                                        str >> pPlayer[0]->human_human >> colon
                                                >> pPlayer[1]->human_human >> colon
                                                >> pPlayer[2]->human_human;
                                    else if (gameTypeValue == "HumanEasy")
                                        str >> pPlayer[0]->human_easy >> colon
                                                >> pPlayer[1]->human_easy >> colon
                                                >> pPlayer[2]->human_easy;
                                    else if (gameTypeValue == "HumanHard")
                                        str >> pPlayer[0]->human_hard >> colon
                                                >> pPlayer[1]->human_hard >> colon
                                                >> pPlayer[2]->human_hard;
                                    else if (gameTypeValue == "EasyEasy")
                                        str >> pPlayer[0]->easy_easy >> colon
                                                >> pPlayer[1]->easy_easy >> colon
                                                >> pPlayer[2]->easy_easy;
                                    else if (gameTypeValue == "EasyHard")
                                        str >> pPlayer[0]->easy_hard >> colon
                                                >> pPlayer[1]->easy_hard >> colon
                                                >> pPlayer[2]->easy_hard;
                                    else if (gameTypeValue == "HardHard")
                                        str >> pPlayer[0]->hard_hard >> colon
                                                >> pPlayer[1]->hard_hard >> colon
                                                >> pPlayer[2]->hard_hard;
                                }
                            }
                        }

                    }
                }
            }
        }
#ifdef LIBXMLCPP_EXCEPTIONS_ENABLED
    }
    catch (const std::exception& ex)
    {
        std::cout << "Exception caught: " << ex.what () << std::endl;
    }
#endif //LIBXMLCPP_EXCEPTIONS_ENABLED

    return statistics;
}

void TStatistics::SaveStatistics (const TStatistics& statistics)
{
    /* $XDG_DATA_HOME defines the base directory relative to which user
     * specific data files should be stored. If $XDG_DATA_HOME is either not
     * set or empty, a default equal to $HOME/.local/share should be used. */

    Glib::ustring homeDir = Background::GetXDGBaseDir ("XDG_DATA_HOME", "/.local/share");

    xmlpp::Document document;
    document.set_internal_subset ("Statistics", "", GNOME_DATADIR "/" + TStatistics::FILENAME_DTD);
    //document.set_internal_subset ("Statistics", "", TStatistics::FILENAME_DTD);

    //Quod is the default namespace prefix.
    //Declares the namespace and uses its prefix for this node
    xmlpp::Element* nodeRoot = document.create_root_node ("Statistics", "http://sourceforge.net/projects/gquod", "Quod");
    nodeRoot->set_namespace_declaration ("http://sourceforge.net/projects/gquod", "Quod"); //Also associate this prefix with this namespace:
    nodeRoot->set_namespace ("Quod");

    xmlpp::Element* nodeChild = NULL,
            *nodePlayer = nodeRoot->add_child ("FirstPlayer");
    nodePlayer->set_attribute ("id", TPLAYER_STRING[BLUE_PLAYER]);

    const int ANY_BLUE[][3] = {
        {statistics.blue_blue.human_human,
            statistics.blue_red.human_human,
            statistics.blue_draw.human_human},
        {statistics.blue_blue.human_easy,
            statistics.blue_red.human_easy,
            statistics.blue_draw.human_easy},
        {statistics.blue_blue.human_hard,
            statistics.blue_red.human_hard,
            statistics.blue_draw.human_hard},
        { statistics.blue_blue.easy_easy,
            statistics.blue_red.easy_easy,
            statistics.blue_draw.easy_easy},
        { statistics.blue_blue.easy_hard,
            statistics.blue_red.easy_hard,
            statistics.blue_draw.easy_hard},
        {statistics.blue_blue.hard_hard,
            statistics.blue_red.hard_hard,
            statistics.blue_draw.hard_hard}
    };

    const int ANY_RED[][3] = {
        {statistics.red_blue.human_human,
            statistics.red_red.human_human,
            statistics.red_draw.human_human},
        {statistics.red_blue.human_easy,
            statistics.red_red.human_easy,
            statistics.red_draw.human_easy},
        {statistics.red_blue.human_hard,
            statistics.red_red.human_hard,
            statistics.red_draw.human_hard},
        { statistics.red_blue.easy_easy,
            statistics.red_red.easy_easy,
            statistics.red_draw.easy_easy},
        { statistics.red_blue.easy_hard,
            statistics.red_red.easy_hard,
            statistics.red_draw.easy_hard},
        {statistics.red_blue.hard_hard,
            statistics.red_red.hard_hard,
            statistics.red_draw.hard_hard}
    };

    for (int i = 0; i < 6; i++)
    {
        nodeChild = nodePlayer->add_child ("GameType");
        nodeChild->set_attribute ("id", TPlayerStat::GAME_TYPE[i]);
        nodeChild->set_child_text (Glib::ustring::compose ("%1:%2:%3",
                                                           ANY_BLUE[i][0],
                                                           ANY_BLUE[i][1],
                                                           ANY_BLUE[i][2]));
    }

    nodePlayer = nodeRoot->add_child ("FirstPlayer");
    nodePlayer->set_attribute ("id", TPLAYER_STRING[RED_PLAYER]);

    for (int i = 0; i < 6; i++)
    {
        nodeChild = nodePlayer->add_child ("GameType");
        nodeChild->set_attribute ("id", TPlayerStat::GAME_TYPE[i]);
        nodeChild->set_child_text (Glib::ustring::compose ("%1:%2:%3",
                                                           ANY_RED[i][0],
                                                           ANY_RED[i][1],
                                                           ANY_RED[i][2]));
    }

#ifndef NDEBUG
    Glib::ustring whole = document.write_to_string_formatted ();

    std::cout << "XML built at runtime: " << std::endl << whole << std::endl;
    std::cout << "default namespace: " << nodeRoot->get_namespace_uri () << std::endl;
#else //NDEBUG
    document.write_to_file_formatted (homeDir + '/' + TStatistics::FILENAME_STATISTICS);
#endif //NDEBUG
}
